from __future__ import annotations

from awepatch.utils import _remove_source_indent  # type: ignore


def test_remove_source_indent_basic() -> None:
    """Test removing common indentation from code lines."""
    lines = [
        "    def foo():\n",
        "        return 42\n",
    ]
    result = _remove_source_indent(lines)
    assert result == [
        "def foo():\n",
        "    return 42\n",
    ]


def test_remove_source_indent_no_indent() -> None:
    """Test with lines that have no indentation."""
    lines = [
        "def foo():\n",
        "    return 42\n",
    ]
    result = _remove_source_indent(lines)
    assert result == [
        "def foo():\n",
        "    return 42\n",
    ]


def test_remove_source_indent_blank_lines() -> None:
    """Test that blank lines are replaced with newline character."""
    lines = [
        "    def foo():\n",
        "        pass\n",
        "    \n",  # Blank line with spaces less than indent
        "        return 42\n",
    ]
    result = _remove_source_indent(lines)
    assert result == [
        "def foo():\n",
        "    pass\n",
        "\n",
        "    return 42\n",
    ]


def test_remove_source_indent_short_blank_line() -> None:
    """Test blank line shorter than indent level."""
    lines = [
        "        def foo():\n",
        "  \n",  # Only 2 spaces, less than 8 space indent
        "            pass\n",
    ]
    result = _remove_source_indent(lines)
    assert result == [
        "def foo():\n",
        "\n",
        "    pass\n",
    ]


def test_remove_source_indent_empty_blank_line() -> None:
    """Test completely empty blank line."""
    lines = [
        "    def foo():\n",
        "\n",  # Empty line
        "        pass\n",
    ]
    result = _remove_source_indent(lines)
    assert result == [
        "def foo():\n",
        "\n",
        "    pass\n",
    ]


def test_remove_source_indent_single_line() -> None:
    """Test with a single line."""
    lines = ["    x = 1\n"]
    result = _remove_source_indent(lines)
    assert result == ["x = 1\n"]


def test_remove_source_indent_deeply_nested() -> None:
    """Test with deeply nested indentation."""
    lines = [
        "            if True:\n",
        "                if True:\n",
        "                    return 1\n",
    ]
    result = _remove_source_indent(lines)
    assert result == [
        "if True:\n",
        "    if True:\n",
        "        return 1\n",
    ]



from __future__ import annotations

import inspect
from functools import partial, wraps
from typing import TYPE_CHECKING, ParamSpec, TypeVar

import pytest

from awepatch import Patch, patch_callable

if TYPE_CHECKING:
    from collections.abc import Callable

_P = ParamSpec("_P")
_R = TypeVar("_R")


def test_patch_function_and_restore() -> None:
    def function_to_patch(x: int) -> int:
        x = x * 2
        return x

    original_result = function_to_patch(5)
    assert original_result == 10

    with patch_callable(
        function_to_patch,
        [Patch("x = x * 2", "x = x * 3", "replace")],
    ):
        patched_result = function_to_patch(5)
        assert patched_result == 15

    restored_result = function_to_patch(5)
    assert restored_result == original_result


def test_patch_function_mode() -> None:
    def fn() -> list[int]:
        a: list[int] = []
        a.append(3)
        return a

    with patch_callable(
        fn,
        [Patch("a.append(3)", "a.append(4)", "replace")],
    ):
        assert fn() == [4]

    with patch_callable(
        fn,
        [Patch("a.append(3)", "a.append(4)", "before")],
    ):
        assert fn() == [4, 3]

    with patch_callable(
        fn,
        [Patch("a.append(3)", "a.append(4)", "after")],
    ):
        assert fn() == [3, 4]


def test_patch_function_with_blank_lines() -> None:
    # do not remove blank lines in `fn` function body
    def fn() -> list[int]:
        a: list[int] = []

        a.append(3)
        return a

    assert (
        inspect.getsource(fn)
        == """    def fn() -> list[int]:
        a: list[int] = []

        a.append(3)
        return a
"""
    )

    with patch_callable(
        fn,
        [Patch("a.append(3)", "a.append(4)", "replace")],
    ):
        assert fn() == [4]


def test_patch_method() -> None:
    class MyClass:
        def method_to_patch(self, z: int) -> int:
            z += 1
            return z

    with patch_callable(
        MyClass.method_to_patch, [Patch("z += 1", "z += 5", "replace")]
    ):
        obj = MyClass()
        assert obj.method_to_patch(10) == 15


def test_patch_obj_method() -> None:
    class MyClass:
        def method_to_patch(self, z: int) -> int:
            z += 1
            return z

    obj = MyClass()
    with patch_callable(obj.method_to_patch, [Patch("z += 1", "z += 5", "replace")]):
        assert obj.method_to_patch(10) == 15


def test_patch_class_method() -> None:
    class MyClass:
        @classmethod
        def class_method_to_patch(cls, z: int) -> int:
            z += 2
            return z

    with patch_callable(
        MyClass.class_method_to_patch, [Patch("z += 2", "z += 10", "replace")]
    ):
        assert MyClass.class_method_to_patch(10) == 20


def test_patch_static_method() -> None:
    class MyClass:
        @staticmethod
        def static_method_to_patch(z: int) -> int:
            z += 3
            return z

    with patch_callable(
        MyClass.static_method_to_patch, [Patch("z += 3", "z += 15", "replace")]
    ):
        assert MyClass.static_method_to_patch(10) == 25


def test_patch_partial_function() -> None:
    from functools import partial

    def function_to_patch(a: int, b: int) -> int:
        return a + b

    partial_func = partial(function_to_patch, 5)

    with patch_callable(
        partial_func, [Patch("return a + b", "return a * b", "replace")]
    ):
        assert partial_func(3) == 15  # 5 * 3


def test_wrapper_function() -> None:
    def wrapper(func: Callable[_P, _R]) -> Callable[_P, _R]:
        @wraps(func)
        def inner(*args: _P.args, **kwargs: _P.kwargs) -> _R:
            return func(*args, **kwargs)

        return inner

    @wrapper
    def function_to_patch(x: int) -> int:
        x += 4
        return x

    with patch_callable(function_to_patch, [Patch("x += 4", "x += 10", "replace")]):
        assert function_to_patch(6) == 16


def test_partialed_wrapped_class_method() -> None:
    def wrapper(func: Callable[_P, _R]) -> Callable[_P, _R]:
        @wraps(func)
        def inner(*args: _P.args, **kwargs: _P.kwargs) -> _R:
            return func(*args, **kwargs)

        return inner

    class MyClass:
        @classmethod
        @wrapper
        def method_to_patch_123(cls, z: int) -> int:
            z += 2
            return z

    method_to_patch = partial(MyClass.method_to_patch_123, 10)
    method_to_patch = wrapper(method_to_patch)

    assert MyClass.method_to_patch_123(10) == 12
    with patch_callable(method_to_patch, [Patch("z += 2", "z += 8", "replace")]):
        assert method_to_patch() == 18


def test_wrapper_and_and_partial_class_method() -> None:
    def wrapper(func: Callable[_P, _R]) -> Callable[_P, _R]:
        @wraps(func)
        def inner(*args: _P.args, **kwargs: _P.kwargs) -> _R:
            return func(*args, **kwargs)

        return inner

    class MyClass:
        @classmethod
        @wrapper
        def method_to_patch(cls, z: int) -> int:
            z += 2
            return z

    assert MyClass.method_to_patch(10) == 12
    with patch_callable(
        MyClass.method_to_patch, [Patch("z += 2", "z += 8", "replace")]
    ):
        assert MyClass.method_to_patch(10) == 18


def test_wrap_by_function() -> None:
    def method_to_patch(z: int) -> int:
        z += 2
        return z

    method_to_patch = pytest.mark.skip("xxx")(method_to_patch)

    assert method_to_patch(10) == 12

    with patch_callable(method_to_patch, [Patch("z += 2", "z += 8", "replace")]):
        assert method_to_patch(10) == 18


def test_lambda_func() -> None:
    lambda_func: Callable[[int], int] = lambda x: x + 5  # noqa

    assert lambda_func(10) == 15

    with (
        pytest.raises(TypeError, match="Cannot patch lambda functions"),
        patch_callable(lambda_func, [Patch("x + 5", "x + 10", "replace")]),
    ):
        pass


def test_patch_async_function() -> None:
    import asyncio

    async def async_function_to_patch(x: int) -> int:
        x = x * 2
        return x

    original_result = asyncio.run(async_function_to_patch(5))
    assert original_result == 10

    with patch_callable(
        async_function_to_patch,
        [Patch("x = x * 2", "x = x * 3", "replace")],
    ):
        patched_result = asyncio.run(async_function_to_patch(5))
        assert patched_result == 15

    restored_result = asyncio.run(async_function_to_patch(5))
    assert restored_result == original_result


def test_callable_patcher_basic() -> None:
    """Test patch_callable basic functionality."""

    def my_function(x: int) -> int:
        return x + 1

    patcher = patch_callable(
        my_function, Patch("return x + 1", "return x + 2", "replace")
    )

    # Test before applying
    assert my_function(3) == 4

    # Apply patch
    patched_func = patcher.apply()
    assert patched_func(3) == 5
    assert my_function(3) == 5  # Original function is also patched

    # Restore
    patcher.restore()
    assert my_function(3) == 4


def test_callable_patcher_context_manager() -> None:
    """Test patch_callable as context manager."""

    def my_function(x: int) -> int:
        return x * 2

    patcher = patch_callable(
        my_function, Patch("return x * 2", "return x * 3", "replace")
    )

    assert my_function(4) == 8

    with patcher:
        assert my_function(4) == 12

    assert my_function(4) == 8


def test_callable_patcher_apply_restore_multiple_times() -> None:
    """Test applying and restoring patches multiple times."""

    def my_function(x: int) -> int:
        return x + 5

    patcher = patch_callable(
        my_function, Patch("return x + 5", "return x + 10", "replace")
    )

    assert my_function(2) == 7

    # Apply, restore, apply again
    patcher.apply()
    assert my_function(2) == 12

    patcher.restore()
    assert my_function(2) == 7

    patcher.apply()
    assert my_function(2) == 12

    patcher.restore()
    assert my_function(2) == 7


def test_callable_patcher_invalid_patch() -> None:
    """Test patch_callable with invalid patch."""

    def my_function(x: int) -> int:
        return x + 1

    with pytest.raises(TypeError, match="patch must be a Patch or a list of Patch"):
        patch_callable(my_function, "not a patch")  # type: ignore[arg-type]

    with pytest.raises(TypeError, match="patch must be a Patch or a list of Patch"):
        patch_callable(my_function, [Patch("a", "b", "replace"), "not a patch"])  # type: ignore[list-item]



from __future__ import annotations

import ast
import atexit
import inspect
import os.path
import re
import shutil
import sys
import tempfile
import threading
from collections import defaultdict
from functools import partial
from types import CodeType, FunctionType
from typing import TYPE_CHECKING, Any, Literal, NamedTuple, TypeAlias

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping, Sequence

Mode: TypeAlias = Literal["before", "after", "replace"]


class Patch(NamedTuple):
    """A single patch operation.

    Attributes:
        pattern: The pattern to search for in the source code.
        repl: The replacement code or AST statements.
        mode: The mode of patching (before/after/replace).

    """

    trgt: str | re.Pattern[str] | tuple[str | re.Pattern[str], ...]
    repl: str | Sequence[ast.stmt]
    mode: Mode = "before"


def get_origin_function(func: Callable[..., Any]) -> Callable[..., Any]:
    """Get the object wrapped by *func*.

    Follows the chain of :attr:`__wrapped__` or :attr:`__func__` attributes or
    :class:`functools.partial` objects, returning the last object in the chain.

    :exc:`ValueError` is raised if a cycle is encountered.

    """
    f = func  # remember the original func for error reporting
    # Memoise by id to tolerate non-hashable objects, but store objects to
    # ensure they aren't destroyed, which would allow their IDs to be reused.
    memo = {id(f): f}
    recursion_limit = sys.getrecursionlimit()
    while not isinstance(func, type):
        if hasattr(func, "__wrapped__"):
            func = func.__wrapped__  # pyright: ignore[reportFunctionMemberAccess]
        elif hasattr(func, "__func__"):
            func = func.__func__  # pyright: ignore[reportFunctionMemberAccess]
        elif isinstance(func, partial):
            func = func.func
        else:
            break
        id_func = id(func)
        if (id_func in memo) or (len(memo) >= recursion_limit):
            raise ValueError(f"wrapper loop when unwrapping {f!r}")
        memo[id_func] = func
    return func


def load_stmts(code: str) -> list[ast.stmt]:
    return ast.parse(code).body


def _remove_source_indent(lines: Sequence[str]) -> list[str]:
    """Remove common leading indentation from code lines."""

    indent = len(lines[0]) - len(lines[0].lstrip())
    # for blank lines indent may be larger than line length, so we replace with '\n'
    lines = [line[indent:] if len(line) > indent else "\n" for line in lines]
    return lines


def get_source_lines(
    obj: CodeType | FunctionType,
) -> list[str]:
    """Get the source lines of a function or code object.

    Args:
        obj (CodeType | FunctionType): The function or code object to get the source
          lines.

    """
    source_lines = inspect.getsourcelines(obj)[0]
    return _remove_source_indent(source_lines)


def _find_function_code(module: CodeType) -> CodeType:
    funcs = [
        func
        for func in module.co_consts
        if isinstance(func, CodeType) and func.co_name != "<lambda>"
    ]
    if len(funcs) != 1:
        raise ValueError("Only single function definitions are supported")
    return funcs[0]


_AWEPATCH_DEBUG = int(os.getenv("AWEPATCH_DEBUG") or 0)


_awepatch_temp_dir: str | None = None
_awepatch_temp_dir_lock = threading.Lock()


def _get_awepatch_temp_dir() -> str:
    """Get or create the temporary directory for awepatch."""
    global _awepatch_temp_dir

    if _awepatch_temp_dir is not None:
        return _awepatch_temp_dir

    with _awepatch_temp_dir_lock:
        if _awepatch_temp_dir is not None:
            return _awepatch_temp_dir

        _awepatch_temp_dir = tempfile.mkdtemp(prefix="awepatch-")
        if not _AWEPATCH_DEBUG:
            # Register cleanup at exit
            atexit.register(
                lambda: (
                    _awepatch_temp_dir
                    and shutil.rmtree(_awepatch_temp_dir, ignore_errors=True)
                )
            )
        return _awepatch_temp_dir


_patch_count = 0
_patch_count_lock = threading.Lock()


def load_function_code(
    func: ast.FunctionDef | ast.AsyncFunctionDef,
    origin: str = "",
) -> CodeType:
    """Load a function's code object from its AST module.

    Args:
        func (ast.Module): The AST module containing the function definition.
        origin (str, optional): The origin location for the function. Defaults to "".
        temp_dir (str): Temporary directory to store the generated file.

    Returns:
        CodeType: The code object of the function.

    """

    source = ast.unparse(func)
    origin = f" (patched from {origin})" if origin else ""
    source = f"# generated by awepatch{origin}\n{source}"

    with _patch_count_lock:
        global _patch_count
        _patch_count += 1

    temp_dir = _get_awepatch_temp_dir()
    file_path = os.path.join(temp_dir, f"patch_{_patch_count:03d}__{func.name}.py")
    with open(file_path, "w") as f:
        f.write(source)
    module_code = compile(source, filename=file_path, mode="exec")
    func_code = _find_function_code(module_code)
    if func_code.co_name != func.name:
        # This should never happen though better to be safe than sorry
        raise ValueError("Function name mismatch!")

    return func_code


def _is_match_node(
    node: ast.AST,
    source: Sequence[str],
    pattern: str | re.Pattern[str],
) -> bool:
    """Check if the AST node matches the target pattern.

    Args:
        node: The AST node to check
        source: The source code lines
        pattern: The target pattern

    Returns:
        True if the AST node matches the target pattern, False otherwise

    """
    if not isinstance(node, ast.stmt):
        return False

    node_lines = source[node.lineno - 1 : node.end_lineno]
    node_source = "".join(_remove_source_indent(node_lines))

    return bool(
        isinstance(pattern, str)
        and node_source.startswith(pattern)
        or isinstance(pattern, re.Pattern)
        and pattern.match(node_source)
    )


def _find_matched_node(
    node: ast.AST,
    source: Sequence[str],
    target: tuple[str | re.Pattern[str], ...],
) -> tuple[ast.AST, str, int] | None:
    """Recursively find the target AST node matching the target patterns.

    Args:
        node: The AST node to search
        source: The source code lines
        target: The target patterns

    Returns:
        The AST node matching the target patterns, or None if not found

    """
    matched: tuple[ast.AST, str, int] | None = None

    for key, field in ast.iter_fields(node):
        if isinstance(field, list):
            for index, item in enumerate(field):  # pyright: ignore[reportUnknownVariableType, reportUnknownArgumentType]
                if not isinstance(item, ast.AST):
                    continue

                if _is_match_node(item, source, target[0]):
                    if len(target) == 1:
                        tmp_matched = (node, key, index)
                    else:
                        tmp_matched = _find_matched_node(item, source, target[1:])
                else:
                    tmp_matched = _find_matched_node(item, source, target)

                if tmp_matched is not None:
                    if matched is not None:
                        raise ValueError(
                            f"Multiple matches found for target pattern {target}"
                        )
                    matched = tmp_matched

        elif isinstance(field, ast.AST):
            if len(target) == 1:
                tmp_matched = _find_matched_node(field, source, target)
            elif _is_match_node(field, source, target[0]):
                tmp_matched = _find_matched_node(field, source, target[1:])
            else:
                tmp_matched = _find_matched_node(field, source, target)

            if tmp_matched is not None:
                if matched is not None:
                    raise ValueError(
                        f"Multiple matches found for target pattern {target}"
                    )
                matched = tmp_matched

    return matched


def _apply_stmts_patches(
    stmts: list[Any],
    index: int,
    patches: dict[Mode, Sequence[ast.stmt]],
) -> int:
    """Insert or replace statements at the specified position with multiple modes.

    Applies patches in order: before, replace, after.

    Returns:
        The total number of statements added (can be negative if replaced with fewer).

    """
    i = index

    # Apply 'before' first
    if "before" in patches:
        stmts[i:i] = patches["before"]
        i += len(patches["before"])  # Adjust index after insertion

    # Apply 'replace' (which removes the original statement)
    if "replace" in patches:
        stmts[i : i + 1] = patches["replace"]
        i += len(patches["replace"]) - 1  # Adjust for replacement

    # Apply 'after' last
    if "after" in patches:
        stmts[i + 1 : i + 1] = patches["after"]
        i += len(patches["after"])  # Adjust index after insertion

    return i - index  # Total number of statements added


def _check_targets(
    target: str | re.Pattern[str] | tuple[str | re.Pattern[str], ...],
) -> tuple[str | re.Pattern[str], ...]:
    """Check the targets of a Patch object.

    Args:
        target: The target pattern(s) to check.

    Raises:
        ValueError: If the targets are invalid.

    """
    if not target:
        raise ValueError("Target pattern cannot be empty")

    if isinstance(target, (str, re.Pattern)):
        return (target,)
    elif (
        isinstance(target, tuple)  # pyright: ignore[reportUnnecessaryIsInstance]
        and all(isinstance(t, (str, re.Pattern)) for t in target)  # pyright: ignore[reportUnnecessaryIsInstance]
    ):
        return target
    else:
        raise TypeError("Unknown target pattern type")


def _compile_patches(
    func: ast.FunctionDef | ast.AsyncFunctionDef,
    source: list[str],
    patches: list[Patch],
) -> defaultdict[tuple[ast.AST, str], defaultdict[int, dict[Mode, Sequence[ast.stmt]]]]:
    """Compile patches into a dictionary mapping line numbers to mode-specific
    statements.

    Args:
        func: The AST function definition
        source: The source code lines of the function
        patches: List of Patch objects to compile

    Returns:
        Dictionary mapping line numbers to dictionaries of mode->statements

    Raises:
        ValueError: If duplicate modes on same line or conflicting patches

    """
    compiled_patches: defaultdict[
        tuple[ast.AST, str], defaultdict[int, dict[Mode, Sequence[ast.stmt]]]
    ] = defaultdict(lambda: defaultdict(dict))

    for patch in patches:
        # Get target line number
        target = _find_matched_node(
            func,
            source,
            _check_targets(patch.trgt),
        )
        if target is None:
            raise ValueError(f"No match found for target pattern {patch.trgt!r}")

        # Prepare replacement statements
        repl_stmts = (
            load_stmts(patch.repl) if isinstance(patch.repl, str) else patch.repl
        )

        if target_patches := compiled_patches[target[:2]][target[2]]:
            # Check for duplicate mode on the same line
            if patch.mode in target_patches:
                raise ValueError(
                    f"Multiple '{patch.mode}' patches on the same target {patch.trgt!r}"
                )
            # Check for conflicting patches: replace cannot be combined with other modes
            if patch.mode == "replace" or "replace" in target_patches:
                raise ValueError(
                    "Cannot combine 'replace' with other modes "
                    f"on target {patch.trgt!r}"
                )
        target_patches[patch.mode] = repl_stmts

    return compiled_patches


def _apply_compiled_patches(
    compiled_patches: Mapping[
        tuple[ast.AST, str], Mapping[int, dict[Mode, Sequence[ast.stmt]]]
    ],
) -> None:
    """Apply compiled patches to the AST function definition.

    Args:
        func: The AST function definition
        compiled_patches: Compiled patches mapping line numbers to mode-specific
          statements

    Returns:
        True if any patches were applied, False otherwise

    """
    for (node, key), index_patches in compiled_patches.items():
        trgt_items = getattr(node, key)
        offset = 0
        for index, patches in sorted(index_patches.items()):
            offset += _apply_stmts_patches(trgt_items, index + offset, patches)


def ast_patch(
    func: CodeType | FunctionType,
    patches: list[Patch],
) -> ast.FunctionDef | ast.AsyncFunctionDef:
    """Patch the AST of a function or code object.

    Args:
        func (CodeType | FunctionType): The function or code object to patch.
        patches (list[Patch]): List of Patch objects for applying multiple patches.

    Returns:
        ast.Module: The modified AST module contains only one function definition.

    """
    # Validate arguments
    if not patches:
        raise ValueError("patches list cannot be empty")

    # 1. Get source code
    source = get_source_lines(func)

    # 2. Parse AST and validate
    func_ast = ast.parse("".join(source))
    if len(func_ast.body) != 1:
        raise ValueError("Only single function definitions are supported")
    func_ast = func_ast.body[0]  # type: ignore[assignment]
    if not isinstance(func_ast, (ast.FunctionDef, ast.AsyncFunctionDef)):
        raise ValueError(f"Not a function definition: {type(func_ast)}")

    # 3. Clear decorators to get plain function code
    func_ast.decorator_list.clear()

    # 4. Pre-compile all patches: find line numbers and prepare replacement statements
    compiled_patches = _compile_patches(func_ast, source, patches)

    # 5. Apply all compiled patches to the AST
    _apply_compiled_patches(compiled_patches)

    return func_ast



from __future__ import annotations

import ast
import re
from dataclasses import dataclass
from typing import Any

import pytest

from awepatch.utils import Patch, ast_patch


def test_ast_patch_function() -> None:
    def function_to_patch(x: int) -> int:
        x = x + 10
        return x

    res_ast_obj = ast_patch(
        function_to_patch, [Patch("x = x + 10", "x = x + 20", "replace")]
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(x: int) -> int:
    x = x + 20
    return x"""
    )


def test_ast_patch_function_dataclass() -> None:
    @dataclass
    class Data:
        value: int
        matched: bool

    def function_to_patch(x: int) -> Data:
        a = Data(
            value=x,
            matched=False,
        )
        return a

    res_ast_obj = ast_patch(function_to_patch, [Patch("a = Data(", "a = x", "after")])
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(x: int) -> Data:
    a = Data(value=x, matched=False)
    a = x
    return a"""
    )


def test_ast_patch_function_complex() -> None:
    @dataclass
    class Data:
        value: int
        matched: bool

    def function_to_patch(x: int) -> list[Data]:
        ll: list[Data] = []
        for i in range(5):
            ll.append(
                Data(
                    value=x + i,
                    matched=False,
                )
            )
        return ll

    res_ast_obj = ast_patch(function_to_patch, [Patch("ll.append(", "a = x", "after")])
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(x: int) -> list[Data]:
    ll: list[Data] = []
    for i in range(5):
        ll.append(Data(value=x + i, matched=False))
        a = x
    return ll"""
    )


def test_ast_patch_function_remove() -> None:
    def function_to_patch(x: int) -> int:
        x = x + 10
        return x

    res_ast_obj = ast_patch(function_to_patch, [Patch("x = x + 10", "", "replace")])
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(x: int) -> int:
    return x"""
    )


def test_ast_patch_one_line_decorator() -> None:
    @classmethod
    def function_to_patch(cls: Any, x: int) -> int:  # type: ignore  # noqa: ANN401
        x = x + 10
        return x

    res_ast_obj = ast_patch(
        function_to_patch,
        [Patch("x = x + 10", "x = x + 20", "replace")],
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(cls: Any, x: int) -> int:
    x = x + 20
    return x"""
    )


def test_ast_patch_function_multi_line_decorator() -> None:
    @pytest.mark.skip(
        reason="remove_decorators not implemented yet",
    )
    def function_to_patch(x: int) -> int:
        x = x + 10
        return x

    res_ast_obj = ast_patch(
        function_to_patch,
        [Patch("x = x + 10", "x = x + 20", "replace")],
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(x: int) -> int:
    x = x + 20
    return x"""
    )


def test_ast_patch_mode_before() -> None:
    """Test inserting code before the target line."""

    def function_to_patch(x: int) -> int:
        x = x + 10
        return x

    res_ast_obj = ast_patch(
        function_to_patch, [Patch("return x", "x = x * 2", "before")]
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(x: int) -> int:
    x = x + 10
    x = x * 2
    return x"""
    )


def test_ast_patch_mode_after() -> None:
    """Test inserting code after the target line."""

    def function_to_patch(x: int) -> int:
        x = x + 10
        return x

    res_ast_obj = ast_patch(
        function_to_patch, [Patch("x = x + 10", "x = x * 2", "after")]
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(x: int) -> int:
    x = x + 10
    x = x * 2
    return x"""
    )


def test_ast_patch_with_regex_pattern() -> None:
    """Test patching using a regex pattern."""

    def function_to_patch(x: int) -> int:
        x = x + 10
        return x

    res_ast_obj = ast_patch(
        function_to_patch,
        [Patch(re.compile(r"x = x \+ \d+"), "x = x + 30", "replace")],
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(x: int) -> int:
    x = x + 30
    return x"""
    )


def test_ast_patch_nested_statements() -> None:
    """Test patching code inside nested statements (if, for, while)."""

    def function_to_patch(x: int) -> int:
        if x > 0:
            x = x + 10
        return x

    res_ast_obj = ast_patch(
        function_to_patch, [Patch("x = x + 10", "x = x + 20", "replace")]
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(x: int) -> int:
    if x > 0:
        x = x + 20
    return x"""
    )


def test_ast_patch_deeply_nested_statements() -> None:
    """Test patching code inside deeply nested statements."""

    def function_to_patch(x: int) -> int:
        if x > 0:
            for _ in range(10):
                while x < 100:
                    x = x + 5
        return x

    res_ast_obj = ast_patch(
        function_to_patch, [Patch("x = x + 5", "x = x + 10", "replace")]
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(x: int) -> int:
    if x > 0:
        for _ in range(10):
            while x < 100:
                x = x + 10
    return x"""
    )


def test_ast_patch_with_multiple_statements() -> None:
    """Test replacing with multiple statements."""

    def function_to_patch(x: int) -> int:
        x = x + 10
        return x

    res_ast_obj = ast_patch(
        function_to_patch,
        [Patch("x = x + 10", "y = x * 2\nx = y + 5", "replace")],
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(x: int) -> int:
    y = x * 2
    x = y + 5
    return x"""
    )


def test_ast_patch_with_ast_statements() -> None:
    """Test patching with AST statement objects instead of string."""

    def function_to_patch(x: int) -> int:
        x = x + 10
        return x

    # Create AST statements directly
    new_stmts = ast.parse("x = x + 50").body

    res_ast_obj = ast_patch(
        function_to_patch, [Patch("x = x + 10", new_stmts, "replace")]
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(x: int) -> int:
    x = x + 50
    return x"""
    )


def test_ast_patch_error_pattern_not_found() -> None:
    """Test error when pattern is not found in the function."""

    def function_to_patch(x: int) -> int:
        x = x + 10
        return x

    with pytest.raises(ValueError, match="No match found for target pattern"):
        ast_patch(function_to_patch, [Patch("x = x + 999", "x = x + 20", "replace")])


def test_ast_patch_error_multiple_matches() -> None:
    """Test error when pattern matches multiple lines."""

    def function_to_patch(x: int) -> int:
        x = x + 10
        x = x + 10
        return x

    with pytest.raises(ValueError, match="Multiple matches found for target pattern"):
        ast_patch(function_to_patch, [Patch("x = x + 10", "x = x + 20", "replace")])


def test_ast_patch_with_try_except() -> None:
    """Test patching code inside try-except blocks."""

    def function_to_patch(x: int) -> int:
        try:
            x = x + 10
        except Exception:
            x = 0
        return x

    res_ast_obj = ast_patch(
        function_to_patch, [Patch("x = x + 10", "x = x + 20", "replace")]
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(x: int) -> int:
    try:
        x = x + 20
    except Exception:
        x = 0
    return x"""
    )


def test_ast_patch_with_context_manager() -> None:
    """Test patching code inside with statements."""

    def function_to_patch(x: int) -> int:
        with open("test.txt") as f:  # type: ignore  # noqa: F841
            x = x + 10
        return x

    res_ast_obj = ast_patch(
        function_to_patch, [Patch("x = x + 10", "x = x + 20", "replace")]
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(x: int) -> int:
    with open('test.txt') as f:
        x = x + 20
    return x"""
    )


def test_ast_patch_multiline_statement() -> None:
    """Test patching a statement that spans multiple lines."""

    def function_to_patch(x: int) -> int:
        result = x + 10 + 20 + 30
        return result

    # Match the multiline statement
    res_ast_obj = ast_patch(
        function_to_patch,
        [Patch("result = x + 10 + 20 + 30", "result = x * 2", "replace")],
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(x: int) -> int:
    result = x * 2
    return result"""
    )


def test_ast_patch_for_loop() -> None:
    """Test patching code inside for loops."""

    def function_to_patch(items: list[int]) -> int:
        total = 0
        for item in items:
            total = total + item
        return total

    res_ast_obj = ast_patch(
        function_to_patch,
        [Patch("total = total + item", "total = total + item * 2", "replace")],
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(items: list[int]) -> int:
    total = 0
    for item in items:
        total = total + item * 2
    return total"""
    )


def test_ast_patch_while_loop() -> None:
    """Test patching code inside while loops."""

    def function_to_patch(x: int) -> int:
        while x < 100:
            x = x + 10
        return x

    res_ast_obj = ast_patch(
        function_to_patch, [Patch("x = x + 10", "x = x + 20", "replace")]
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(x: int) -> int:
    while x < 100:
        x = x + 20
    return x"""
    )


def test_ast_patch_match_statement() -> None:
    """Test patching code inside match statements (Python 3.10+)."""

    def function_to_patch(x: int) -> str:
        match x:
            case 1:
                result = "one"
            case _:
                result = "other"
        return result

    res_ast_obj = ast_patch(
        function_to_patch, [Patch('result = "one"', 'result = "ONE"', "replace")]
    )
    res_str = ast.unparse(res_ast_obj)

    # ast.unparse may use single quotes instead of double quotes
    assert "result = 'ONE'" in res_str or 'result = "ONE"' in res_str
    assert "match x:" in res_str


def test_ast_patch_insert_multiple_lines_before() -> None:
    """Test inserting multiple lines before a statement."""

    def function_to_patch(x: int) -> int:
        return x

    res_ast_obj = ast_patch(
        function_to_patch,
        [Patch("return x", "x = x * 2\nx = x + 10", "before")],
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(x: int) -> int:
    x = x * 2
    x = x + 10
    return x"""
    )


def test_ast_patch_insert_multiple_lines_after() -> None:
    """Test inserting multiple lines after a statement."""

    def function_to_patch(x: int) -> int:
        x = x + 5
        return x

    res_ast_obj = ast_patch(
        function_to_patch,
        [Patch("x = x + 5", "x = x * 2\nx = x + 10", "after")],
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(x: int) -> int:
    x = x + 5
    x = x * 2
    x = x + 10
    return x"""
    )


def test_ast_patch_complex_function() -> None:
    """Test patching a more complex function with multiple control structures."""

    def function_to_patch(items: list[int]) -> int:
        total = 0
        for item in items:
            if item > 0:
                try:  # noqa: SIM105
                    total = total + item
                except Exception:
                    pass
        return total

    res_ast_obj = ast_patch(
        function_to_patch,
        [Patch("total = total + item", "total = total + item * 10", "replace")],
    )
    res_str = ast.unparse(res_ast_obj)

    assert "total = total + item * 10" in res_str
    assert "for item in items:" in res_str
    assert "if item > 0:" in res_str


def test_ast_patch_list_comprehension_in_body() -> None:
    """Test patching when function contains list comprehensions."""

    def function_to_patch(items: list[int]) -> list[int]:
        result = [x * 2 for x in items]
        return result

    res_ast_obj = ast_patch(
        function_to_patch,
        [
            Patch(
                "result = [x * 2 for x in items]",
                "result = [x * 3 for x in items]",
                "replace",
            )
        ],
    )
    res_str = ast.unparse(res_ast_obj)

    assert "x * 3" in res_str


def test_ast_patch_lambda_in_body() -> None:
    """Test patching when function contains lambda expressions."""

    def function_to_patch(x: int) -> int:
        func = lambda n: n + 10  # pyright: ignore # noqa: E731
        result = func(x)  # type: ignore
        return result  # type: ignore

    res_ast_obj = ast_patch(
        function_to_patch,
        [Patch("result = func(x)  # type: ignore", "result = func(x) * 2", "replace")],
    )
    res_str = ast.unparse(res_ast_obj)

    assert "result = func(x) * 2" in res_str


def test_multiple_patches_different_lines() -> None:
    """Test applying multiple patches to different lines."""

    def function_to_patch(x: int) -> int:
        x = x + 10
        x = x * 2
        return x

    res_ast_obj = ast_patch(
        function_to_patch,
        [
            Patch("x = x + 10", "x = x + 20", "replace"),
            Patch("x = x * 2", "x = x * 3", "replace"),
        ],
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(x: int) -> int:
    x = x + 20
    x = x * 3
    return x"""
    )


def test_multiple_patches_before_and_after_same_line() -> None:
    """Test applying both before and after patches to the same line."""

    def function_to_patch(x: int) -> int:
        x = x + 10
        return x

    res_ast_obj = ast_patch(
        function_to_patch,
        [
            Patch("x = x + 10", "print('before return')", "before"),
            Patch("x = x + 10", "print('after return')", "after"),
        ],
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(x: int) -> int:
    print('before return')
    x = x + 10
    print('after return')
    return x"""
    )


def test_multiple_patches_mixed_modes() -> None:
    """Test applying patches with different modes to different lines."""

    def function_to_patch(x: int, y: int) -> int:
        x = x + 10
        y = y * 2
        result = x + y
        return result

    res_ast_obj = ast_patch(
        function_to_patch,
        [
            Patch("x = x + 10", "print('processing x')", "before"),
            Patch("y = y * 2", "y = y * 3", "replace"),
            Patch("result = x + y", "print('result calculated')", "after"),
        ],
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(x: int, y: int) -> int:
    print('processing x')
    x = x + 10
    y = y * 3
    result = x + y
    print('result calculated')
    return result"""
    )


def test_multiple_patches_with_default_mode() -> None:
    """Test applying multiple patches with default 'before' mode."""

    def function_to_patch(x: int) -> int:
        x = x + 10
        x = x * 2
        return x

    res_ast_obj = ast_patch(
        function_to_patch,
        [
            Patch("x = x + 10", "print('before first')"),  # default mode='before'
            Patch("x = x * 2", "print('before second')"),
        ],
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(x: int) -> int:
    print('before first')
    x = x + 10
    print('before second')
    x = x * 2
    return x"""
    )


def test_multiple_patches_error_duplicate_mode_same_line() -> None:
    """Test error when multiple patches with same mode target the same line."""

    def function_to_patch(x: int) -> int:
        x = x + 10
        return x

    with pytest.raises(
        ValueError, match="Multiple 'before' patches on the same target"
    ):
        ast_patch(
            function_to_patch,
            [
                Patch("x = x + 10", "print('first')", "before"),
                Patch("x = x + 10", "print('second')", "before"),
            ],
        )


def test_multiple_patches_error_replace_conflict() -> None:
    """Test error when replace is combined with other modes on the same line."""

    def function_to_patch(x: int) -> int:
        x = x + 10
        return x

    with pytest.raises(ValueError, match="Cannot combine 'replace' with other modes"):
        ast_patch(
            function_to_patch,
            [
                Patch("x = x + 10", "x = x + 20", "replace"),
                Patch("x = x + 10", "print('before')", "before"),
            ],
        )


def test_multiple_patches_nested_and_toplevel() -> None:
    """Test applying patches to both nested and top-level statements."""

    def function_to_patch(x: int) -> int:
        if x > 0:
            x = x + 10
        x = x * 2
        return x

    res_ast_obj = ast_patch(
        function_to_patch,
        [
            Patch("x = x + 10", "x = x + 20", "after"),
            Patch("x = x * 2", "x = x * 3", "replace"),
            Patch("return x", "print('done')", "before"),
        ],
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(x: int) -> int:
    if x > 0:
        x = x + 10
        x = x + 20
    x = x * 3
    print('done')
    return x"""
    )


def test_multiple_patches_many_lines() -> None:
    """Test applying many patches to a larger function."""

    def function_to_patch(items: list[int]) -> int:
        total = 0
        count = 0
        for item in items:
            if item > 0:
                total = total + item
                count = count + 1
        return total

    res_ast_obj = ast_patch(
        function_to_patch,
        [
            Patch("total = 0", "print('initializing total')", "after"),
            Patch("total = total + item", "total = total + item * 2", "replace"),
            Patch("count = count + 1", "print(f'count: {count}')", "after"),
            Patch("return total", "print(f'final total: {total}')", "before"),
        ],
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(items: list[int]) -> int:
    total = 0
    print('initializing total')
    count = 0
    for item in items:
        if item > 0:
            total = total + item * 2
            count = count + 1
            print(f'count: {count}')
    print(f'final total: {total}')
    return total"""
    )


def test_multiline_multiple_patches() -> None:
    """Test applying patches to both nested and top-level statements."""

    def function_to_patch(x: int) -> int:
        if x > 0:
            x = x + 10
        x = x * 2
        return x

    res_ast_obj = ast_patch(
        function_to_patch,
        [
            Patch(
                "x = x + 10", "x = x + 20\nx = x + 20\nx = x + 20\nx = x + 20", "after"
            ),
            Patch("x = x * 2", "x = x * 3", "after"),
        ],
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""def function_to_patch(x: int) -> int:
    if x > 0:
        x = x + 10
        x = x + 20
        x = x + 20
        x = x + 20
        x = x + 20
    x = x * 2
    x = x * 3
    return x"""
    )


def test_async_function_patch() -> None:
    """Test patching an async function."""

    async def function_to_patch(x: int) -> int:
        x = x + 10
        return x

    res_ast_obj = ast_patch(
        function_to_patch, [Patch("x = x + 10", "x = x + 20", "replace")]
    )
    res_str = ast.unparse(res_ast_obj)

    assert (
        res_str
        == r"""async def function_to_patch(x: int) -> int:
    x = x + 20
    return x"""
    )



# type: ignore
# ruff: noqa
from __future__ import annotations

import functools
import sys
import unittest

from awepatch.utils import get_origin_function


class NTimesUnwrappable:
    def __init__(self, n):
        self.n = n
        self._next = None

    @property
    def __wrapped__(self):
        if self.n <= 0:
            raise Exception("Unwrapped too many times")
        if self._next is None:
            self._next = NTimesUnwrappable(self.n - 1)
        return self._next


class TestUnwrap(unittest.TestCase):
    def test_unwrap_one(self):
        def func(a, b):
            return a + b

        wrapper = functools.lru_cache(maxsize=20)(func)
        self.assertIs(get_origin_function(wrapper), func)

    def test_unwrap_several(self):
        def func(a, b):
            return a + b

        wrapper = func
        for __ in range(10):

            @functools.wraps(wrapper)
            def wrapper():
                pass

        self.assertIsNot(wrapper.__wrapped__, func)
        self.assertIs(get_origin_function(wrapper), func)

    def test_cycle(self):
        def func1():
            pass

        func1.__wrapped__ = func1
        with self.assertRaisesRegex(ValueError, "wrapper loop"):
            get_origin_function(func1)

        def func2():
            pass

        func2.__wrapped__ = func1
        func1.__wrapped__ = func2
        with self.assertRaisesRegex(ValueError, "wrapper loop"):
            get_origin_function(func1)
        with self.assertRaisesRegex(ValueError, "wrapper loop"):
            get_origin_function(func2)

    def test_unhashable(self):
        def func():
            pass

        func.__wrapped__ = None

        class C:
            __hash__ = None
            __wrapped__ = func

        self.assertIsNone(get_origin_function(C()))

    def test_recursion_limit(self):
        obj = NTimesUnwrappable(sys.getrecursionlimit() + 1)
        with self.assertRaisesRegex(ValueError, "wrapper loop"):
            get_origin_function(obj)

    def test_wrapped_descriptor(self):
        self.assertIs(get_origin_function(NTimesUnwrappable), NTimesUnwrappable)
        self.assertIs(get_origin_function(staticmethod), staticmethod)
        self.assertIs(get_origin_function(classmethod), classmethod)
        self.assertIs(get_origin_function(staticmethod(classmethod)), classmethod)
        self.assertIs(get_origin_function(classmethod(staticmethod)), staticmethod)



from __future__ import annotations

import ast
import re
from typing import TYPE_CHECKING

import pytest

from awepatch.utils import _find_matched_node  # pyright: ignore[reportPrivateUsage]

if TYPE_CHECKING:
    from collections.abc import Sequence


def test_find_matched_node_exact_string_match() -> None:
    """Test finding a node by exact string match."""
    source = [
        "def foo():\n",
        "    x = 1\n",
        "    y = 2\n",
        "    return x + y\n",
    ]
    tree = ast.parse("".join(source))
    matched = _find_matched_node(tree.body[0], source, ("x = 1",))
    assert matched is not None
    assert ast.unparse(getattr(matched[0], matched[1])[matched[2]]) == "x = 1"


def test_find_matched_node_in_nested_block() -> None:
    """Test finding a node inside a nested control flow block."""

    source_code: str = """def example_function(x):
    if x > 0:
        x = x * 3
    x = x * 2
    return x
"""

    tree = ast.parse(source_code)
    target = ("x = x * 3",)
    source_lines: Sequence[str] = source_code.splitlines(keepends=True)

    matched = _find_matched_node(
        tree.body[0],
        source_lines,
        target,
    )

    assert matched is not None
    assert ast.unparse(getattr(matched[0], matched[1])[matched[2]]) == "x = x * 3"


def test_find_matched_node_with_context() -> None:
    """Test disambiguating duplicate patterns using context lines."""

    source_code: str = """def example_function(x):
    if x > 0:
        x = x * 2
    x = x * 2
    return x
"""

    tree = ast.parse(source_code)

    target = ("if x > 0:", "x = x * 2")

    source_lines: Sequence[str] = source_code.splitlines(keepends=True)
    matched = _find_matched_node(
        tree.body[0],
        source_lines,
        target,
    )

    assert matched is not None
    assert ast.unparse(matched[0]) == "if x > 0:\n    x = x * 2"
    assert ast.unparse(getattr(matched[0], matched[1])[matched[2]]) == "x = x * 2"


def test_find_matched_node_raises_on_ambiguous_match() -> None:
    """Test that multiple matches raise ValueError when no context is provided."""

    source_code: str = """def example_function(x):
    if x > 0:
        x = x * 2
    x = x * 2
    return x
"""

    tree = ast.parse(source_code)
    target = ("x = x * 2",)
    source_lines: Sequence[str] = source_code.splitlines(keepends=True)

    with pytest.raises(ValueError, match="Multiple matches found for target pattern"):
        _find_matched_node(
            tree.body[0],
            source_lines,
            target,
        )


def test_find_matched_node_regex_pattern() -> None:
    """Test finding a node using a compiled regex pattern."""
    source = [
        "def foo():\n",
        "    x = 1\n",
        "    y = 2\n",
        "    return x + y\n",
    ]
    pattern = re.compile(r"x = \d+")
    tree = ast.parse("".join(source))
    matched = _find_matched_node(tree.body[0], source, (pattern,))
    assert matched is not None
    assert ast.unparse(getattr(matched[0], matched[1])[matched[2]]) == "x = 1"


def test_find_matched_node_ignores_trailing_whitespace() -> None:
    """Test that matching ignores trailing whitespace in source lines."""
    source = [
        "def foo():\n",
        "    x = 1    \n",
        "    y = 2\n",
        "    return x + y\n",
    ]
    tree = ast.parse("".join(source))
    matched = _find_matched_node(tree.body[0], source, ("x = 1",))
    assert matched is not None
    assert ast.unparse(getattr(matched[0], matched[1])[matched[2]]) == "x = 1"


def test_find_matched_node_returns_none_when_not_found() -> None:
    """Test that None is returned when the pattern doesn't match any node."""
    source = [
        "def foo():\n",
        "    x = 1\n",
        "    y = 2\n",
        "    return x + y\n",
    ]
    tree = ast.parse("".join(source))
    matched = _find_matched_node(tree.body[0], source, ("z = 3",))
    assert matched is None


def test_find_matched_node_raises_on_duplicate_lines() -> None:
    """Test that identical duplicate lines raise ValueError without context."""
    source = [
        "def foo():\n",
        "    x = 1\n",
        "    x = 1\n",
        "    return x\n",
    ]
    tree = ast.parse("".join(source))
    with pytest.raises(ValueError, match="Multiple matches found for target pattern"):
        _find_matched_node(tree.body[0], source, ("x = 1",))


def test_find_matched_node_complex_regex_pattern() -> None:
    """Test finding a node using a complex regex with multiple groups."""
    source = [
        "def foo():\n",
        "    result = calculate(10, 20, 30)\n",
        "    return result\n",
    ]
    tree = ast.parse("".join(source))
    pattern = re.compile(r"result = calculate\(\d+,\s*\d+,\s*\d+\)")
    matched = _find_matched_node(tree.body[0], source, (pattern,))
    assert matched is not None
    assert (
        ast.unparse(getattr(matched[0], matched[1])[matched[2]])
        == "result = calculate(10, 20, 30)"
    )


def test_find_matched_node_case_sensitive_matching() -> None:
    """Test that pattern matching is case-sensitive."""
    source = [
        "def foo():\n",
        "    X = 1\n",
        "    return X\n",
    ]
    tree = ast.parse("".join(source))
    matched = _find_matched_node(tree.body[0], source, ("x = 1",))
    assert matched is None



from __future__ import annotations

import ast
from typing import TYPE_CHECKING

from awepatch.utils import get_source_lines, load_stmts

if TYPE_CHECKING:
    from collections.abc import Callable


def test_load_stmts_single_statement() -> None:
    """Test loading a single statement."""
    code = "x = 1"
    stmts = load_stmts(code)
    assert len(stmts) == 1
    assert isinstance(stmts[0], ast.Assign)


def test_load_stmts_multiple_statements() -> None:
    """Test loading multiple statements."""
    code = "x = 1\ny = 2\nz = 3"
    stmts = load_stmts(code)
    assert len(stmts) == 3
    assert all(isinstance(stmt, ast.Assign) for stmt in stmts)


def test_load_stmts_complex_statements() -> None:
    """Test loading complex statements."""
    code = """
if x > 0:
    y = 1
else:
    y = 2
"""
    stmts = load_stmts(code)
    assert len(stmts) == 1
    assert isinstance(stmts[0], ast.If)


def test_load_stmts_with_imports() -> None:
    """Test loading import statements."""
    code = "import os\nfrom pathlib import Path"
    stmts = load_stmts(code)
    assert len(stmts) == 2
    assert isinstance(stmts[0], ast.Import)
    assert isinstance(stmts[1], ast.ImportFrom)


def test_load_stmts_with_function_def() -> None:
    """Test loading function definition."""
    code = """
def foo(x):
    return x * 2
"""
    stmts = load_stmts(code)
    assert len(stmts) == 1
    assert isinstance(stmts[0], ast.FunctionDef)
    assert stmts[0].name == "foo"


def test_load_stmts_with_class_def() -> None:
    """Test loading class definition."""
    code = """
class MyClass:
    def __init__(self):
        pass
"""
    stmts = load_stmts(code)
    assert len(stmts) == 1
    assert isinstance(stmts[0], ast.ClassDef)
    assert stmts[0].name == "MyClass"


def test_get_source_lines_function() -> None:
    """Test getting source lines from a function."""

    def sample_function(x: int) -> int:
        y = x + 10
        return y

    lines = get_source_lines(sample_function)
    assert len(lines) > 0
    assert lines[0] == "def sample_function(x: int) -> int:\n"


def test_get_source_lines_nested_function() -> None:
    """Test getting source lines from a nested function."""

    def outer() -> Callable[[int], int]:
        def inner(x: int) -> int:
            return x * 2

        return inner

    inner_func = outer()
    lines = get_source_lines(inner_func)
    assert len(lines) > 0
    assert "def inner" in lines[0]


def test_get_source_lines_removes_common_indentation() -> None:
    """Test that get_source_lines removes common leading indentation."""

    class MyClass:
        def method(self, x: int) -> int:
            return x + 1

    lines = get_source_lines(MyClass.method)
    # First line should not have leading spaces
    assert lines[0].startswith("def method")


def test_load_stmts_empty_code() -> None:
    """Test loading empty code returns empty list."""
    code = ""
    stmts = load_stmts(code)
    assert len(stmts) == 0


def test_load_stmts_only_whitespace() -> None:
    """Test loading code with only whitespace."""
    code = "   \n\n   "
    stmts = load_stmts(code)
    assert len(stmts) == 0


def test_load_stmts_with_comments() -> None:
    """Test that comments are ignored when loading statements."""
    code = """
# This is a comment
x = 1  # Inline comment
# Another comment
y = 2
"""
    stmts = load_stmts(code)
    # Comments are not statements
    assert len(stmts) == 2
    assert all(isinstance(stmt, ast.Assign) for stmt in stmts)



from __future__ import annotations

import ast
from typing import TYPE_CHECKING

from awepatch.utils import get_source_lines, load_function_code

if TYPE_CHECKING:
    from collections.abc import Callable


def hello() -> int:
    return 4


def test_load_func_code_with_default() -> None:
    def fn_with_default(y: int, x: int = hello()) -> int:
        return x + y

    fn_with_default_ast = ast.parse("".join(get_source_lines(fn_with_default)))

    fn_with_default_co = load_function_code(fn_with_default_ast.body[0])  # pyright: ignore[reportArgumentType]
    assert fn_with_default_co.co_name == fn_with_default.__name__
    fn_with_default.__code__ = fn_with_default_co
    assert fn_with_default(1) == 5


def test_load_func_code_with_nested() -> None:
    def fn_with_nested(x: int = 5) -> int:
        def _nested() -> int:
            return 5 + x

        return _nested()

    fn_with_nested_ast = ast.parse("".join(get_source_lines(fn_with_nested)))

    fn_with_nested_co = load_function_code(fn_with_nested_ast.body[0])  # pyright: ignore[reportArgumentType]
    assert fn_with_nested_co.co_name == fn_with_nested.__name__
    fn_with_nested.__code__ = fn_with_nested_co
    assert fn_with_nested(10) == 15


def test_load_func_code_with_lambda() -> None:
    def fn_with_lambda(x: Callable[[], int] = lambda: 5, y: int = 10) -> int:
        return x() + y

    fn_with_lambda_ast = ast.parse("".join(get_source_lines(fn_with_lambda)))

    fn_with_lambda_co = load_function_code(fn_with_lambda_ast.body[0])  # pyright: ignore[reportArgumentType]
    assert fn_with_lambda_co.co_name == fn_with_lambda.__name__
    fn_with_lambda.__code__ = fn_with_lambda_co
    assert fn_with_lambda(y=9) == 14
    assert fn_with_lambda(x=lambda: 3) == 13



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from awepatch._version import __commit_id__, __version__, __version_tuple__
from awepatch.utils import (
    Patch,
    ast_patch,
    get_origin_function,
    load_function_code,
)

if TYPE_CHECKING:
    from collections.abc import Callable
    from types import CodeType, TracebackType


def _get_patched_code(
    func: Callable[..., Any],
    patches: list[Patch],
) -> CodeType:
    """Get the patched code object of a function without modifying it.

    Args:
        func (Callable[..., Any]): The function to patch.
        patches (list[Patch]): List of Patch objects for applying multiple patches.

    Returns:
        The patched code object

    """
    # Patch the function's AST
    patched_func_ast = ast_patch(func.__code__, patches)
    patched_func_code = load_function_code(
        patched_func_ast,
        origin=f"{func.__code__.co_filename}:{func.__code__.co_firstlineno}",
    )
    return patched_func_code


def _check_callable(func: Callable[..., Any]) -> Callable[..., Any]:
    """Check if the provided object is a valid callable for patching and return the
    original function.

    Args:
        func: The callable to check.

    Returns:
        The original function if valid.

    """
    if not callable(func):
        raise TypeError(f"Expected a function, got: {type(func)}")
    func = get_origin_function(func)
    if not callable(func):
        raise TypeError(f"Expected a function, got: {type(func)}")
    if func.__name__ == "<lambda>":
        raise TypeError("Cannot patch lambda functions")
    return func


def _check_patches(patch: Any) -> list[Patch]:  # noqa: ANN401
    """Check and normalize the patches input.

    Args:
        patch: A single Patch or a list of Patch objects.

    Returns:
        A list of Patch objects.

    """
    if isinstance(patch, Patch):
        return [patch]
    elif isinstance(patch, list) and all(isinstance(p, Patch) for p in patch):  # pyright: ignore[reportUnknownVariableType]
        return patch  # pyright: ignore[reportUnknownVariableType]
    else:
        raise TypeError("patch must be a Patch or a list of Patch objects")


class CallablePatcher(ABC):
    @abstractmethod
    def apply(self) -> Callable[..., Any]:
        """Apply the patches to the function."""

    @abstractmethod
    def restore(self) -> None:
        """Restore the original function."""

    def __enter__(self) -> Callable[..., Any]:
        """Enter the context manager, applying the patches."""
        return self.apply()

    def __exit__(
        self,
        exc_type: type[BaseException],
        exc_value: BaseException,
        traceback: TracebackType,
    ) -> None:
        """Exit the context manager, restoring the original function."""
        self.restore()


class _CallablePatcher(CallablePatcher):
    """A class for patching callables using AST manipulation."""

    def __init__(
        self,
        func: Callable[..., Any],
        patch: Patch | list[Patch],
    ) -> None:
        """Initialize a CallablePatcher.

        The CallablePatcher can be used as a context manager to apply patches
        to a callable function.

        Args:
            func (Callable[..., Any]): The function to patch.
            patch (Patch | list[Patch]): Patch or list of Patch objects for applying
                multiple patches.

        """
        func = _check_callable(func)
        patches = _check_patches(patch)

        self._func = func
        self._original_code = func.__code__
        self._patched_code = _get_patched_code(func, patches)

    def apply(self) -> Callable[..., Any]:
        """Apply the patches to the function."""
        self._func.__code__ = self._patched_code
        return self._func

    def restore(self) -> None:
        """Restore the original function."""
        self._func.__code__ = self._original_code


def patch_callable(
    func: Callable[..., Any],
    /,
    patch: Patch | list[Patch],
) -> CallablePatcher:
    """Patch a callable using AST manipulation.

    Args:
        func (Callable[..., Any]): The function to patch.
        patch (Patch | list[Patch]): Patch or list of Patch objects for applying
            multiple patches.

    Examples:
        >>> from awepatch import Patch, patch_callable
        >>> def my_function(x):
        ...     return x + 1
        >>> with patch_callable(my_function, Patch("x + 1", "x + 2", "replace")):
        ...     assert my_function(3) == 5

    """

    return _CallablePatcher(func, patch)


__all__ = (
    "__commit_id__",
    "__version__",
    "__version_tuple__",
    "CallablePatcher",
    "Patch",
    "patch_callable",
)
